import re
import uuid
from urllib.parse import urljoin, urlparse

from django.conf import settings
from django.contrib.contenttypes.fields import GenericRelation
from django.contrib.gis.db.models.functions import SnapToGrid
from django.contrib.postgres.constraints import ExclusionConstraint
from django.contrib.postgres.fields import HStoreField, RangeOperators
from django.contrib.postgres.indexes import GinIndex
from django.core.exceptions import ValidationError
from django.core.validators import MaxValueValidator, MinValueValidator, RegexValidator
from django.db import connections, models, transaction
from django.db.models import Count, Deferrable, Q, Window
from django.db.models.functions import Cast, Least
from django.urls import reverse
from django.utils.functional import cached_property
from enumfields import Enum, EnumField

import pgtrigger
from arkindex.documents.dates import InterpretedDateMixin
from arkindex.documents.deletion import delete_element
from arkindex.documents.managers import CorpusManager, ElementManager
from arkindex.project.aws import S3FileMixin
from arkindex.project.default_corpus import DEFAULT_CORPUS_TYPES
from arkindex.project.fields import ArrayConcat, ArrayField, LinearRingField
from arkindex.project.models import IndexableModel
from arkindex.users.models import Role


class Corpus(IndexableModel):
    """
    A group of elements, usually linked to users or projects
    """
    id = models.UUIDField(default=uuid.uuid4, primary_key=True, editable=False)
    name = models.CharField(max_length=250)
    description = models.TextField(default="")
    memberships = GenericRelation("users.Right", "content_id")
    # Restrict top level elements to a certain type.
    # Efficient on large corpora as the type filter is applied first, assuming most elements of this type have no parent.
    top_level_type = models.ForeignKey(
        "documents.ElementType",
        on_delete=models.SET_NULL,
        related_name="top_level_corpora",
        null=True,
        blank=True
    )

    # Is this corpus publicly readable ?
    public = models.BooleanField(default=False)

    # Is this corpus indexable ?
    indexable = models.BooleanField(default=False)

    # Specific manager for ACL
    objects = CorpusManager()

    class Meta:
        verbose_name_plural = "corpora"

    def __str__(self):
        return self.name

    def get_absolute_url(self):
        return urljoin(
            settings.PUBLIC_HOSTNAME,
            reverse("frontend-corpus-details", kwargs={"pk": self.id}),
        )

    def create_default_types(self):
        self.types.bulk_create(
            ElementType(corpus=self, **values)
            for values in DEFAULT_CORPUS_TYPES
        )

    def is_writable(self, user) -> bool:
        """
        Whether a user has write access to this corpus
        """
        if user.is_anonymous or getattr(user, "is_agent", False):
            return False
        if user.is_admin:
            return True
        from arkindex.users.utils import get_max_level
        level = get_max_level(user, self)
        return level is not None and level >= Role.Contributor.value

    def is_processable(self, user) -> bool:
        """
        Whether a user can create and execute processes on this corpus
        """
        if user.is_anonymous or getattr(user, "is_agent", False):
            return False
        if user.is_admin:
            return True
        from arkindex.users.utils import get_max_level
        level = get_max_level(user, self)
        return level is not None and level >= Role.Admin.value


class ElementType(models.Model):
    id = models.UUIDField(default=uuid.uuid4, primary_key=True, editable=False)
    corpus = models.ForeignKey(Corpus, related_name="types", on_delete=models.CASCADE)
    slug = models.SlugField(db_index=False)
    display_name = models.CharField(max_length=250)
    folder = models.BooleanField(default=False)
    indexable = models.BooleanField(default=False)
    color = models.CharField(
        max_length=6,
        default="28b62c",
        validators=[
            RegexValidator(
                "^[0-9a-f]{6}$",
                flags=re.IGNORECASE,
                message="Color should be a hexadecimal color code"
            )
        ]
    )

    class Meta:
        constraints = [
            models.UniqueConstraint(fields=["corpus", "slug"], name="corpus_unique_type_slug"),
        ]

    def __str__(self):
        return self.display_name


class ElementPath(models.Model):
    """
    Rolling cache table, with full paths exposed for each element
    An element can have multiple paths up to the top
    """
    id = models.UUIDField(default=uuid.uuid4, primary_key=True, editable=False)
    element = models.ForeignKey("documents.Element", related_name="paths", on_delete=models.CASCADE)
    path = ArrayField(
        models.UUIDField(),
    )
    ordering = models.PositiveIntegerField(default=0)

    class Meta:
        indexes = [
            GinIndex(fields=["path"]),
            models.Index("path__last", name="element_path_last"),
        ]
        constraints = [
            models.UniqueConstraint(fields=("element", "path"), name="unique_element_paths"),
            # Allow either empty paths or non-empty paths on the same element, but not both at once
            ExclusionConstraint(
                name="unique_top_level",
                expressions=[
                    # For the same element…
                    ("element", RangeOperators.EQUAL),
                    # Exclude paths that would both have a zero and a non-zero length.
                    # This would mean using (array_length(path, 1) > 0) WITH <>, but <> is not implemented
                    # in GiST indexes for booleans. Instead, we'll use LEAST(array_length(path, 1), 1), an integer:
                    # when the length is zero, this returns 0 because 0<1, when it's non-zero, it returns one.
                    (Least("path__len", 1), RangeOperators.NOT_EQUAL),
                ],
                deferrable=Deferrable.DEFERRED,
            ),
        ]
        triggers = [
            # An element cannot be in two positions within the same parent, but its paths can be duplicated
            # as there might be multiple paths to handle multiple grandparents on the same parent.
            # We therefore cannot use a unique constraint. We could use an exclusion constraint to forbid
            # different orderings, but that requires a GiST index which is heavy and slow. If the query planner
            # decides to use it instead of the GIN index, this can result in severe slowdowns in element lists
            # or element deletions. We use a deferrable trigger instead, which behaves very similarly to how
            # PostgreSQL implements foreign key constraints.
            # After one row is inserted or updated, we run a query to see if there is another path
            # for the same element on the same parent that has a different ordering.
            # If this element has two orderings on the same parent, then the trigger raises an exception.
            pgtrigger.Trigger(
                name="unique_element_orderings",
                operation=pgtrigger.Insert | pgtrigger.Update,
                when=pgtrigger.After,
                level=pgtrigger.Row,
                timing=pgtrigger.Deferred,
                func="""
                IF EXISTS (
                    SELECT 1
                    FROM documents_elementpath p
                    WHERE
                        p.path && NEW.path
                        AND p.path[array_length(p.path, 1)] = NEW.path[array_length(NEW.path, 1)]
                        AND p.element_id = NEW.element_id
                        AND p.ordering <> NEW.ordering
                ) THEN
                    RAISE EXCEPTION 'Each element may only have one ordering within the same parent';
                END IF;
                RETURN NULL;
                """,
            ),
            # Within the same parent, there cannot be two elements with the same ordering,
            # but since paths can be duplicated, there can be the same element twice with the same ordering.
            # We again use another trigger, but with a different condition.
            # After one row is inserted or updated, we run a query to see if there is another path
            # with the same parent and the same ordering but a different element ID.
            # If there is another element in this parent with that same ordering, then the trigger raises an exception.
            pgtrigger.Trigger(
                name="unique_parent_orderings",
                operation=pgtrigger.Insert | pgtrigger.Update,
                when=pgtrigger.After,
                level=pgtrigger.Row,
                timing=pgtrigger.Deferred,
                func="""
                IF EXISTS (
                    SELECT 1
                    FROM documents_elementpath p
                    WHERE
                        p.path && NEW.path
                        AND p.path[array_length(p.path, 1)] = NEW.path[array_length(NEW.path, 1)]
                        AND p.ordering = NEW.ordering
                        AND p.element_id <> NEW.element_id
                ) THEN
                    RAISE EXCEPTION 'Each element within a parent must have a distinct ordering';
                END IF;
                RETURN NULL;
                """,
            ),
        ]


class Element(IndexableModel):
    """
    aka Unité Codicologique
    """
    corpus = models.ForeignKey("documents.Corpus", related_name="elements", on_delete=models.CASCADE)
    # `type` will replace `type` once all of Arkindex switches to generic element types
    type = models.ForeignKey(ElementType, related_name="elements", on_delete=models.PROTECT)
    name = models.CharField(max_length=250)

    creator = models.ForeignKey(
        "users.User",
        related_name="elements",
        on_delete=models.SET_NULL,
        null=True,
        blank=True,
    )
    worker_version = models.ForeignKey(
        "process.WorkerVersion",
        on_delete=models.SET_NULL,
        related_name="elements",
        null=True,
        blank=True,
    )

    worker_run = models.ForeignKey(
        "process.WorkerRun",
        # WorkerRuns that are on processes that have not yet run can be deleted easily by the user, to allow them
        # to customize the worker runs before starting the process.
        # Since WorkerRuns are related to each worker result, deleting a worker run with models.CASCADE
        # could cause a dozen useless SQL queries to look for related classifications and other worker results that
        # could be related to it, and could fill up the RAM.
        # Using models.RESTRICT or PROTECT would also cause Django to check for related classifications to raise a
        # RestrictedError or ProtectedError, which both would cause a HTTP 500.  Using models.DO_NOTHING lets PostgreSQL
        # do the checks itself, which it always does anyway, and crash with an IntegrityError, which also causes a 500.
        on_delete=models.DO_NOTHING,
        related_name="elements",
        null=True,
        blank=True,
    )

    image = models.ForeignKey(
        "images.Image",
        on_delete=models.CASCADE,
        related_name="elements",
        null=True,
        blank=True,
    )

    # SRID 0 does not exist, but means 'No SRID'. A projection is not relevant to us
    # since we are working only with cartesian coordinates.
    polygon = LinearRingField(srid=0, null=True, blank=True)

    rotation_angle = models.SmallIntegerField(
        validators=[MinValueValidator(0), MaxValueValidator(359)],
        default=0,
        help_text="Clockwise rotation to apply to the image after cropping, in degrees.",
    )
    mirrored = models.BooleanField(
        default=False,
        help_text="Mirror the image along the vertical axis before rotating.",
    )

    confidence = models.FloatField(
        validators=[MinValueValidator(0), MaxValueValidator(1)],
        null=True,
        blank=True,
    )

    objects = ElementManager()

    class Meta:
        constraints = [
            # Require either both an image and a polygon, or neither of them
            # Django's implementation of __isnull does not work like most field lookups,
            # making it impossible to use it in this case, especially for __isnull=False.
            # We instead check that either both fields are equal to NULL, or none of the two fields are equal to NULL.
            models.CheckConstraint(
                check=Q(image_id=None, polygon=None) | ~(Q(image_id=None) | Q(polygon=None)),
                name="element_image_and_polygon",
            ),
            models.CheckConstraint(
                check=Q(polygon=None) | Q(polygon__isclosed=True),
                name="element_polygon_closed",
            ),
            models.CheckConstraint(
                check=Q(polygon=None) | Q(polygon=SnapToGrid("polygon", 1)),
                name="element_polygon_integer_coordinates",
            ),
            # Require at least 3+1 points for a triangle.
            models.CheckConstraint(
                check=Q(polygon=None) | Q(polygon__numpoints__gte=4),
                name="element_polygon_size",
            ),
            # Using a PositiveSmallIntegerField for rotation_angle would implicitly add a CHECK (rotation_angle >= 0),
            # but the MaxValueValidator would be ignored, so we use SmallIntegerField and our own constraint for [0, 359].
            models.CheckConstraint(
                check=Q(rotation_angle__gte=0, rotation_angle__lte=359),
                name="element_rotation_angle",
            ),
            # Require at least one of (creator, worker_version) to be null.
            # We need to allow both to be null because both fields were not here since the beginning,
            # so we do not always know who actually created an element.
            models.CheckConstraint(
                check=Q(creator_id=None) | Q(worker_version_id=None),
                name="element_creator_nand_worker_version",
            ),
            # Prevent elements from having a worker run and no worker version
            models.CheckConstraint(
                check=Q(worker_run_id__isnull=True) | Q(worker_version_id__isnull=False),
                name="element_worker_run_requires_worker_version",
            ),
        ]

    def delete(self, *args, **kwargs):
        """
        Use raw SQL for element deletion and bypass Django's cascading
        """
        assert self.id, "Cannot delete an element without an ID"
        delete_element(self.id)
        Element.objects.filter(id=self.id)._raw_delete(using="default")

    @transaction.atomic
    def add_parent(self, parent, skip_children=False):
        """
        Add an element as ancestor
        """
        assert isinstance(parent, Element)
        assert self.corpus_id == parent.corpus_id, "Both elements must be in the same corpus"

        with connections["default"].cursor() as cursor:
            # Fetch some values that we will need to detect which queries to run, and build them:
            # - Whether or not the child has any paths, to detect edge cases
            # - The first non-top-level path the child has, if there's any
            # - Whether the child is already a descendant of this parent, meaning there is nothing to do
            # - Whether the parent is a descendant of the child, which would create a cycle
            # - How many parents the parent has
            # - The first path of any ElementPath of the parent
            # - The next available ordering for direct children of the parent
            # This can almost entirely be expressed using a QuerySet, but Django is stubborn
            # and always tries to add a GROUP BY documents_elementpath.id, which returns
            # wildly incorrect results.
            # We use two CTEs here to make this already fast query twice as fast.
            # The parent_paths CTE is sorted to avoid intermittent test failures.
            cursor.execute(
                """
                WITH parent_paths AS (
                    SELECT path FROM documents_elementpath
                    WHERE element_id = %(parent_id)s
                    ORDER BY path
                ), child_paths AS (
                    SELECT path FROM documents_elementpath
                    WHERE element_id = %(self_id)s
                )
                SELECT
                    EXISTS (SELECT 1 FROM child_paths),
                    (SELECT path FROM child_paths WHERE path <> '{}'::uuid[] LIMIT 1),
                    EXISTS (SELECT 1 FROM child_paths WHERE path @> ARRAY[%(parent_id)s]),
                    EXISTS (SELECT 1 FROM parent_paths WHERE path @> ARRAY[%(self_id)s]),
                    (SELECT COUNT(*) FROM parent_paths),
                    (SELECT path FROM parent_paths LIMIT 1),
                    (
                        SELECT COALESCE(MAX(ordering) + 1, 0)
                        FROM documents_elementpath
                        WHERE path @> ARRAY[%(parent_id)s]
                        AND path[array_length(path, 1)] = %(parent_id)s
                    )
                """,
                {
                    "self_id": self.id,
                    "parent_id": parent.id,
                },
            )
            (
                has_paths,
                existing_parent_path,
                already_has_parent,
                causes_cycle,
                parent_paths_count,
                first_parent_path,
                ordering,
            ) = cursor.fetchone()

            # If the child we are trying to add is among the parents of the new parent,
            # then it would cause a cycle; an element cannot be its own ancestor.
            if causes_cycle:
                raise ValueError("Cannot add a descendant as a parent")

            # If the parent already is among the element's ancestors, then there is nothing to do
            if already_has_parent:
                return

            # Edge case: when the parent has no paths at all, create a top-level path automatically
            if not parent_paths_count:
                parent.paths.create(path=[], ordering=0)
                # Update our state accordingly
                parent_paths_count = 1
                first_parent_path = []

            # Add or update parent paths for this element first.
            # When the child element has other parents, or when the child element has no paths (not even top-level),
            # we only need to insert new paths.
            # Otherwise, when the parent has multiple grandparents, we will need to both insert and update.
            if existing_parent_path or not has_paths or parent_paths_count > 1:
                # This could easily be done using a regular bulk_create, but this make this method a lot harder to test on
                # with assertExactQueries because Django's bulk_create can completely randomize the order in which parent
                # paths are sent in the SQL query.
                query = """
                INSERT INTO documents_elementpath (id, element_id, path, ordering)
                SELECT
                    gen_random_uuid(),
                    %(self_id)s,
                    path || %(parent_id)s,
                    %(ordering)s
                FROM documents_elementpath
                WHERE element_id = %(parent_id)s
                """

                # If the child element has some paths, but no other parents, and the parent has multiple paths,
                # we will UPDATE for one of the paths and INSERT only the rest.
                if has_paths and not existing_parent_path and parent_paths_count > 1:
                    query += "AND path <> %(first_parent_path)s"

                cursor.execute(
                    query,
                    {
                        "self_id": self.id,
                        "parent_id": parent.id,
                        "ordering": ordering,
                        "first_parent_path": first_parent_path,
                    }
                )

            # When the child element only has a top-level path, we update this path to be one of the paths of the new parent
            if has_paths and not existing_parent_path:
                self.paths.filter(path=[]).update(path=first_parent_path + [parent.id], ordering=ordering)

            if skip_children:
                return

            # When there are no other parents, we need to update all the existing paths of the descendants.
            # If there is more than one grandparent (the parent has more than one path), then we will update with
            # the first grandparent, then insert for all the other grandparents as if there were other parents.
            if not existing_parent_path:
                ElementPath.objects.filter(path__overlap=[self.id]).update(
                    path=ArrayConcat(first_parent_path + [parent.id], models.F("path")),
                )

            # When there are other parents, we can build one INSERT query to create new paths for all the descendants.
            # When there were no other parents but the new parent has multiple paths, we can INSERT the remaining paths.
            if existing_parent_path or parent_paths_count > 1:
                # We will take any existing path on the parent element, then look for all existing paths
                # of the descendants of the current element that have this path as a prefix.
                # Filtering by a path prefix means we will include all descendants recursively.
                # We can then remove the prefix to remove the existing parent.
                # For direct children, we will end up with empty arrays.
                # We will then JOIN those suffixes to all the paths of the new parent to build all
                # the new paths to create for all child elements using concatenations.

                query = """
                INSERT INTO documents_elementpath (id, element_id, path, ordering)
                SELECT
                    gen_random_uuid(),
                    child_paths.element_id,
                    new_parent_paths.path || %(new_parent_id)s || child_paths.path[%(start_index)s:],
                    child_paths.ordering
                FROM
                    documents_elementpath child_paths,
                    documents_elementpath new_parent_paths
                WHERE
                    child_paths.path @> %(child_path_prefix)s
                    AND new_parent_paths.element_id = %(new_parent_id)s
                """

                # The prefix to use to look for all the existing children paths is built with
                # an existing path + the ID of the current element, to make sure we do not match
                # the paths of another element.
                # Note that in the final query, we will use @> and not a "starts with" operator,
                # because PostgreSQL does not have an optimized operator for that.  This should not
                # cause any issue as long as the graph is not corrupted, as we should not be seeing
                # anything before the prefix (that would imply some elements are only partially in
                # a parent).

                if existing_parent_path:
                    child_path_prefix = [*existing_parent_path, self.id]
                else:
                    # When we are inserting for the remaining grandparents after the UPDATE done above,
                    # we will add a filter to the WHERE clause to avoid inserting the paths that we just
                    # built using the UPDATE, and use the updated path as the prefix.
                    query += "AND new_parent_paths.path <> %(existing_parent_path)s"
                    existing_parent_path = first_parent_path
                    child_path_prefix = [*first_parent_path, parent.id, self.id]

                cursor.execute(
                    query,
                    {
                        "self_id": self.id,
                        "new_parent_id": parent.id,
                        "existing_parent_path": existing_parent_path,
                        "child_path_prefix": child_path_prefix,
                        # The index at which the existing child path should be cut off.
                        # Since PostgreSQL arrays are 1-based, this will not cut off the entire prefix,
                        # but leave just the last item of the prefix: the current element, which we
                        # actually want to keep.
                        "start_index": len(child_path_prefix),
                    }
                )

    def get_next_order(self):
        """
        Find the next ordering for a new child in ElementPath
        Uses the primary database to avoid stale reads and duplicate orderings,
        and path__overlap to let Postgres use the GIN index before filtering by last item.
        Using the GIN index before path__last is quicker than using path__last alone.
        """
        return (
            ElementPath
            .objects
            .using("default")
            .filter(path__overlap=[self.id], path__last=self.id)
            .aggregate(max=models.Max("ordering") + 1)["max"] or 0
        )

    @transaction.atomic
    def remove_child(self, child):
        """
        Dissociate a child element
        """
        assert isinstance(child, Element)

        with connections["default"].cursor() as cursor:
            # Fetch three values that we will need to detect which queries to run, and build them:
            # - How many paths this child has that have this element as a parent,
            #   which will tell us whether or not we have anything to do and
            #   whether or not this parent has multiple grandparents
            # - How many paths this child has that do not have this element as a parent,
            #   to tell if the child has other parents
            # - Any path on the child that ends with this element's ID, in case we need to do an UPDATE on it.
            # This can almost entirely be expressed using a QuerySet, but Django is stubborn
            # and always tries to add a GROUP BY documents_elementpath.id, which returns
            # wildly incorrect results.
            cursor.execute(
                """
                SELECT
                    COUNT(*) FILTER (WHERE path[array_length(path, 1)] = %(self_id)s),
                    COUNT(*) FILTER (WHERE path[array_length(path, 1)] <> %(self_id)s),
                    (
                        SELECT path
                        FROM documents_elementpath
                        WHERE
                            element_id = %(child_id)s
                            AND path[array_length(path, 1)] = %(self_id)s
                        LIMIT 1
                    )
                FROM documents_elementpath
                WHERE element_id = %(child_id)s
                """,
                {
                    "self_id": self.id,
                    "child_id": child.id,
                },
            )
            parent_paths_count, other_paths_count, first_parent_path = cursor.fetchone()

            # Check child is a direct child of this element
            if not parent_paths_count:
                return

            assert first_parent_path, "At least 1 path to this parent was counted, but no path was retrieved."
            assert first_parent_path[-1] == self.id, "The retrieved path does not match this parent's ID"

            if not other_paths_count:
                # When this element does not have any other parent, we will need to run an UPDATE query
                # to update one of this parent's paths instead of deleting it.  We will strip the path
                # from the child's and the descendants', meaning the child will have a top-level path
                # left (empty array) and the descendants will remain descendants of this child.
                # As an extra precaution, we will check that the path really starts with this prefix before updating,
                # since @> is really a set operation ([1,2,3,4] @> [3,1,4] is true).
                prefix = first_parent_path + [child.id]
                cursor.execute(
                    """
                    UPDATE documents_elementpath
                    SET path = path[%(prefix_size)s:]
                    WHERE
                        path @> %(prefix)s
                        AND path[0:%(prefix_size)s] = %(prefix)s
                    """,
                    {
                        "prefix": prefix,
                        "prefix_size": len(prefix),
                    },
                )

                child.paths.filter(path=first_parent_path).update(path=[])

            if other_paths_count or parent_paths_count > 1:
                # When this element has another parent, we only need to delete the paths to this parent
                # without having to run any update.  When this element both has no other parent, but this
                # parent has multiple grandparents, then once the UPDATE above is performed, we only need
                # to delete all the other paths.
                # We will make two DELETE queries, to remove the child's path to the parent
                # and all of the paths of the child's descendants that include the parent's paths.

                # We start with the child's descendants, using a raw SQL query that will look for the
                # child's paths containing the parent (all paths on the child ending in `self.id`)
                # and delete all paths that start with those paths + the child's ID.
                cursor.execute(
                    """
                    DELETE FROM documents_elementpath child_paths
                    USING documents_elementpath old_parent_paths
                    WHERE
                        old_parent_paths.element_id = %(child_id)s
                        AND old_parent_paths.path[array_length(old_parent_paths.path, 1)] = %(self_id)s
                        AND child_paths.path @> (old_parent_paths.path || %(child_id)s)
                    """,
                    {
                        "self_id": self.id,
                        "child_id": child.id,
                    }
                )

                # Now that the child's descendants are handled, we can clean up the child's own paths.
                child.paths.filter(path__last=self.id).delete()

    @transaction.atomic
    def remove_children(self):
        """
        Remove this parent element from all of its children at once
        """
        # Fetch two values that we will need to detect which queries to run, and build them:
        # - How many paths this parent has, so we can tell if we need to delete paths and not only update them
        # - One path on the parent, so that we can perform the update.
        # In the rare edge case where the element has zero paths, this returns nothing, so we'll act as if there was a top-level path.
        first_parent_path, parent_paths_count = (
            self.paths
            .using("default")
            .annotate(count=Window(Count("*")))
            .values_list("path", "count")
            # Unnecessary for this algorithm to work, but simplifies unit testing a lot.
            .order_by("id")
            .first()
        ) or ([], 0)

        with connections["default"].cursor() as cursor:
            if parent_paths_count > 1:
                # Delete all child paths that are not the first parent path.
                # If we tried to also update those, we would end up with duplicates.
                cursor.execute(
                    """
                    DELETE FROM documents_elementpath child_paths
                    USING documents_elementpath parent_paths
                    WHERE
                        parent_paths.element_id = %(parent_id)s
                        AND parent_paths.path <> %(first_parent_path)s
                        AND child_paths.path @> (parent_paths.path || %(parent_id)s)
                    """,
                    {
                        "parent_id": self.id,
                        "first_parent_path": first_parent_path,
                    },
                )

            # For children that have other parents, delete all the paths for this parent.
            # The paths for the other parents will preserve the structure of the child's descendants,
            # so we will have no updates to make.
            cursor.execute(
                """
                DELETE FROM documents_elementpath parent_paths
                USING documents_elementpath other_paths
                WHERE
                    parent_paths.path && ARRAY[%(parent_id)s]
                    AND other_paths.element_id = parent_paths.element_id
                    AND NOT other_paths.path && ARRAY[%(parent_id)s]
                """,
                {"parent_id": self.id},
            )

            # For the child elements that had no other parent, we have one path starting with `first_parent_path`.
            # We strip that from the children's paths and their descendants',
            # meaning each child will have a top-level path left (empty array)
            # and its descendants will remain descendants of this child.
            # As an extra precaution, we will check that the path really starts with this prefix before updating,
            # since @> is really a set operation ([1,2,3,4] @> [3,1,4] is true).
            prefix = first_parent_path + [self.id]
            # TODO: In Django 5.1, rewrite this without raw SQL by using F() slicing
            # See https://docs.djangoproject.com/en/dev/ref/models/expressions/#slicing-f-expressions
            cursor.execute(
                """
                UPDATE documents_elementpath
                SET path = path[%(prefix_size)s + 1:]
                WHERE
                    path @> %(prefix)s
                    AND path[:%(prefix_size)s] = %(prefix)s
                """,
                {
                    "prefix": prefix,
                    "prefix_size": len(prefix),
                },
            )

    @cached_property
    def thumbnail(self):
        from arkindex.images.models import Thumbnail  # Prevent circular imports
        return Thumbnail(self)

    @property
    def iiif_url(self):
        """
        Cropped IIIF URL for this element's image, if there is one.
        """
        if not self.polygon or not self.image_id:
            return

        from arkindex.project.tools import bounding_box
        x, y, width, height = bounding_box(self.polygon)

        # Handle IIIF URLs with query parameters (cannot use urljoin as it will remove the query parameters)
        parsed_url = urlparse(self.image.url)
        if parsed_url.query:
            return f"{self.image.url}/{x},{y},{width},{height}/full/0/default.jpg"

        return urljoin(
            self.image.url + "/",
            f"{x},{y},{width},{height}/full/0/default.jpg"
        )

    @property
    def iiif_thumbnail_url(self):
        """
        Same as `iiif_url`, but resized to up to 400 pixels and with rotation and mirroring
        """
        if not self.polygon or not self.image_id:
            return

        from arkindex.project.tools import bounding_box
        x, y, width, height = bounding_box(self.polygon)

        rotation_param = str(self.rotation_angle)
        if self.mirrored:
            rotation_param = f"!{rotation_param}"

        # Do no attempt to upsize small images for thumbnails
        thumbnail_height = min(400, height)

        # Handle IIIF URLs with query parameters (cannot use urljoin as it will remove the query parameters)
        parsed_url = urlparse(self.image.url)
        if parsed_url.query:
            return f"{self.image.url}/{x},{y},{width},{height}/,{thumbnail_height}/{rotation_param}/default.jpg"

        return urljoin(
            self.image.url + "/",
            f"{x},{y},{width},{height}/,{thumbnail_height}/{rotation_param}/default.jpg"
        )

    def __str__(self):
        return f"{self.type.display_name}: {self.name}"


class EntityType(models.Model):
    id = models.UUIDField(default=uuid.uuid4, primary_key=True, editable=False)
    name = models.CharField(max_length=250)
    color = models.CharField(
        max_length=6,
        default="ff0000",
        validators=[
            RegexValidator(
                "^[0-9a-f]{6}$",
                flags=re.IGNORECASE,
                message="Color should be a hexadecimal color code"
            )
        ]
    )
    corpus = models.ForeignKey(Corpus, related_name="entity_types", on_delete=models.CASCADE)

    class Meta:
        unique_together = (
            ("name", "corpus"),
        )

    def __str__(self):
        return self.name


class Entity(models.Model):
    """
    Semantic object in arkindex
    """

    id = models.UUIDField(default=uuid.uuid4, primary_key=True, editable=False)
    name = models.TextField()
    type = models.ForeignKey(EntityType, related_name="entities", on_delete=models.DO_NOTHING)
    corpus = models.ForeignKey(Corpus, related_name="entities", on_delete=models.CASCADE)
    metas = HStoreField(null=True, blank=True)
    validated = models.BooleanField(default=False)
    moderator = models.ForeignKey(
        "users.User",
        related_name="entities",
        on_delete=models.SET_NULL,
        null=True,
        blank=True
    )
    worker_version = models.ForeignKey(
        "process.WorkerVersion",
        on_delete=models.CASCADE,
        related_name="entities",
        null=True,
        blank=True,
    )
    worker_run = models.ForeignKey(
        "process.WorkerRun",
        # WorkerRuns that are on processes that have not yet run can be deleted easily by the user, to allow them
        # to customize the worker runs before starting the process.
        # Since WorkerRuns are related to each worker result, deleting a worker run with models.CASCADE
        # could cause a dozen useless SQL queries to look for related classifications and other worker results that
        # could be related to it, and could fill up the RAM.
        # Using models.RESTRICT or PROTECT would also cause Django to check for related classifications to raise a
        # RestrictedError or ProtectedError, which both would cause a HTTP 500.  Using models.DO_NOTHING lets PostgreSQL
        # do the checks itself, which it always does anyway, and crash with an IntegrityError, which also causes a 500.
        on_delete=models.DO_NOTHING,
        related_name="entities",
        null=True,
        blank=True,
    )

    class Meta:
        verbose_name_plural = "Entities"
        constraints = [
            # There can be a worker run ID only if there is a worker version ID,
            # but there can be a worker version ID without a worker run ID (backwards compatibility).
            # In other words, either the worker run ID is null, or the worker version ID is not null.
            models.CheckConstraint(
                check=Q(worker_run_id__isnull=True) | Q(worker_version_id__isnull=False),
                name="entity_worker_run_requires_worker_version",
            )
        ]
        indexes = [
            # ListCorpusEntities can get quite slow when listing millions of entities because it has to sort
            # by name, and there is no index on the name.  It also always filters by corpus ID and falls back to
            # ordering by ID when two names are equal, so we add all of the fields to the index to let Postgres
            # do a single index scan to find all the relevant entities.
            models.Index(fields=["corpus_id", "name", "id"], name="entity_list_index"),
        ]

    def __str__(self):
        return self.name


class TextOrientation(Enum):
    HorizontalLeftToRight = "horizontal-lr"
    HorizontalRightToLeft = "horizontal-rl"
    VerticalRightToLeft = "vertical-rl"
    VerticalLeftToRight = "vertical-lr"


class Transcription(models.Model):
    """
    A transcription on:
     * a zone on an image
    """
    id = models.UUIDField(default=uuid.uuid4, primary_key=True, editable=False)
    element = models.ForeignKey(
        Element,
        on_delete=models.CASCADE,
        related_name="transcriptions",
    )
    worker_version = models.ForeignKey(
        "process.WorkerVersion",
        on_delete=models.CASCADE,
        related_name="transcriptions",
        null=True,
        blank=True,
    )
    worker_run = models.ForeignKey(
        "process.WorkerRun",
        # Equivalent to RESTRICT or PROTECT, but PostgreSQL will do the errors itself
        # instead of Django making extra SQL queries.
        on_delete=models.DO_NOTHING,
        related_name="transcriptions",
        null=True,
        blank=True,
    )
    text = models.TextField()
    orientation = EnumField(TextOrientation, default=TextOrientation.HorizontalLeftToRight, max_length=50)
    confidence = models.FloatField(null=True, blank=True)
    entities = models.ManyToManyField(
        Entity,
        through="TranscriptionEntity",
        related_name="transcriptions",
    )

    class Meta:
        constraints = [
            # There can be a worker run ID only if there is a worker version ID,
            # but there can be a worker version ID without a worker run ID (backwards compatibility).
            # In other words, either the worker run ID is null, or the worker version ID is not null.
            models.CheckConstraint(
                check=Q(worker_version_id__isnull=False) | Q(worker_run_id__isnull=True),
                name="transcription_worker_run_requires_worker_version",
            ),
        ]

    def __str__(self):
        return f"Transcription: {self.text[:20]}"


class TranscriptionEntity(models.Model):
    """
    Many-to-many relationship between Transcription and Entity
    """
    id = models.UUIDField(default=uuid.uuid4, primary_key=True, editable=False)
    transcription = models.ForeignKey(
        Transcription,
        on_delete=models.CASCADE,
        related_name="transcription_entities",
    )
    entity = models.ForeignKey(
        Entity,
        on_delete=models.CASCADE,
        related_name="transcription_entities",
    )
    offset = models.PositiveIntegerField()
    length = models.PositiveIntegerField(validators=[MinValueValidator(1)])

    worker_version = models.ForeignKey(
        "process.WorkerVersion",
        on_delete=models.CASCADE,
        related_name="transcription_entities",
        null=True,
        blank=True,
    )

    worker_run = models.ForeignKey(
        "process.WorkerRun",
        related_name="transcription_entities",
        # WorkerRuns that are on processes that have not yet run can be deleted easily by the user, to allow them
        # to customize the worker runs before starting the process.
        # Since WorkerRuns are related to each worker result, deleting a worker run with models.CASCADE
        # could cause a dozen useless SQL queries to look for related classifications and other worker results that
        # could be related to it, and could fill up the RAM.
        # Using models.RESTRICT or PROTECT would also cause Django to check for related classifications to raise a
        # RestrictedError or ProtectedError, which both would cause a HTTP 500.  Using models.DO_NOTHING lets PostgreSQL
        # do the checks itself, which it always does anyway, and crash with an IntegrityError, which also causes a 500.
        on_delete=models.DO_NOTHING,
        null=True,
        blank=True
    )

    confidence = models.FloatField(validators=[MinValueValidator(0), MaxValueValidator(1)], null=True, blank=True)

    class Meta:
        constraints = [
            # There can be a worker run ID only if there is a worker version ID,
            # but there can be a worker version ID without a worker run ID (backwards compatibility).
            # In other words, either the worker run ID is null, or the worker version ID is not null.
            models.CheckConstraint(
                check=Q(worker_version_id__isnull=False) | Q(worker_run_id__isnull=True),
                name="transcription_entity_worker_run_requires_worker_version",
            ),
            # Add unicity for manual and non manual transcription entities on a transcription.
            # Two constraints are required as Null values are not compared for unicity.
            models.UniqueConstraint(
                fields=["transcription", "entity", "offset", "length"],
                name="transcription_entity_unique_manual",
                condition=Q(worker_version_id__isnull=True),
            ),
            # Keep the old unique constraint on worker versions for transcription entities
            # without worker runs for backwards compatibility
            models.UniqueConstraint(
                fields=["transcription", "entity", "offset", "length", "worker_version"],
                name="transcription_entity_unique_worker_version",
                condition=Q(worker_version_id__isnull=False, worker_run_id__isnull=True),
            ),
            models.UniqueConstraint(
                fields=["transcription", "entity", "offset", "length", "worker_run"],
                name="transcription_entity_unique_worker_run",
                condition=Q(worker_run_id__isnull=False),
            ),
        ]

    def clean(self):
        if self.transcription.element.corpus_id != self.entity.corpus_id:
            raise ValidationError("Transcription and Entity are not in the same corpus")
        if self.offset + self.length > len(self.transcription.text):
            raise ValidationError("Entity position overflows text size")
        super().clean()


class MLClass(models.Model):
    """
    A type of classification
    """
    id = models.UUIDField(default=uuid.uuid4, primary_key=True, editable=False)
    name = models.CharField(max_length=1024)
    corpus = models.ForeignKey("documents.Corpus", on_delete=models.CASCADE, related_name="ml_classes")

    class Meta:
        unique_together = (
            ("name", "corpus"),
        )
        verbose_name_plural = "classes"
        ordering = ("corpus", "name")

    def __str__(self):
        return self.name


class ClassificationState(Enum):
    Pending = "pending"
    Validated = "validated"
    Rejected = "rejected"


class Classification(models.Model):
    """
    A result of a classifier on a page
    """
    id = models.UUIDField(default=uuid.uuid4, primary_key=True, editable=False)
    element = models.ForeignKey(
        Element,
        on_delete=models.CASCADE,
        related_name="classifications",
    )
    worker_version = models.ForeignKey(
        "process.WorkerVersion",
        on_delete=models.CASCADE,
        related_name="classifications",
        null=True,
        blank=True,
    )
    worker_run = models.ForeignKey(
        "process.WorkerRun",
        # WorkerRuns that are on processes that have not yet run can be deleted easily by the user, to allow them
        # to customize the worker runs before starting the process.
        # Since WorkerRuns are related to each worker result, deleting a worker run with models.CASCADE
        # could cause a dozen useless SQL queries to look for related classifications and other worker results that
        # could be related to it, and could fill up the RAM.
        # Using models.RESTRICT or PROTECT would also cause Django to check for related classifications to raise a
        # RestrictedError or ProtectedError, which both would cause a HTTP 500.  Using models.DO_NOTHING lets PostgreSQL
        # do the checks itself, which it always does anyway, and crash with an IntegrityError, which also causes a 500.
        on_delete=models.DO_NOTHING,
        related_name="classifications",
        null=True,
        blank=True,
    )
    moderator = models.ForeignKey(
        "users.User",
        on_delete=models.SET_NULL,
        null=True,
        blank=True,
        related_name="classifications"
    )
    ml_class = models.ForeignKey(
        MLClass,
        on_delete=models.CASCADE,
        related_name="classifications"
    )
    # Predicted class is considered as correct by its creator
    high_confidence = models.BooleanField(default=False)
    state = EnumField(ClassificationState, max_length=16, default=ClassificationState.Pending)
    confidence = models.FloatField()

    class Meta:
        constraints = [
            # There can be a worker run ID only if there is a worker version ID,
            # but there can be a worker version ID without a worker run ID (backwards compatibility).
            # In other words, either the worker run ID is null, or the worker version ID is not null.
            models.CheckConstraint(
                check=Q(worker_version_id__isnull=False) | Q(worker_run_id__isnull=True),
                name="classification_worker_run_requires_worker_version",
            ),
            # Add class unicity for manual and non manual classifications on an element
            # Two constraints are required as Null values are not compared for unicity
            models.UniqueConstraint(
                fields=["element", "ml_class"],
                name="classification_unique_manual",
                condition=Q(worker_version_id__isnull=True),
            ),
            # Keep the old unique constraint on worker versions for classifications without worker runs
            # for backwards compatibility
            models.UniqueConstraint(
                fields=["element", "ml_class", "worker_version"],
                name="classification_unique_worker_version",
                condition=Q(worker_version_id__isnull=False, worker_run_id__isnull=True),
            ),
            models.UniqueConstraint(
                fields=["element", "ml_class", "worker_run"],
                name="classification_unique_worker_run",
                condition=Q(worker_run_id__isnull=False),
            ),
        ]


class MetaType(Enum):
    Text = "text"
    Markdown = "markdown"
    Date = "date"
    Location = "location"
    # Element's original structure reference (intended to be indexed)
    Reference = "reference"
    Numeric = "numeric"
    URL = "url"


class AllowedMetaData(models.Model):
    """
    Defines standard metadata's (type, name) couples for a corpus
    """
    id = models.UUIDField(default=uuid.uuid4, primary_key=True, editable=False)
    corpus = models.ForeignKey("documents.Corpus", related_name="allowed_metadatas", on_delete=models.CASCADE)
    type = EnumField(MetaType, max_length=50)
    name = models.CharField(max_length=250)

    class Meta:
        ordering = ("corpus", "type", "name")
        unique_together = (
            ("corpus", "type", "name"),
        )


class MetaData(InterpretedDateMixin, models.Model):
    """
    Metadatas for elements
    """
    id = models.UUIDField(default=uuid.uuid4, primary_key=True, editable=False)
    element = models.ForeignKey(Element, related_name="metadatas", on_delete=models.CASCADE)
    name = models.CharField(max_length=250)
    type = EnumField(MetaType, max_length=50)
    value = models.TextField()
    entity = models.ForeignKey(Entity, null=True, blank=True, related_name="metadatas", on_delete=models.SET_NULL)
    worker_version = models.ForeignKey(
        "process.WorkerVersion",
        on_delete=models.CASCADE,
        related_name="metadatas",
        null=True,
        blank=True,
    )
    worker_run = models.ForeignKey(
        "process.WorkerRun",
        related_name="metadatas",
        on_delete=models.DO_NOTHING,
        null=True,
        blank=True
    )

    class Meta:
        ordering = ("element", "name", "id")
        constraints = [
            # Either the metadata is not numeric, or it has a value that casts to a double properly.
            # Casting to a double can cause an exception to occur if the value is invalid, so instead we check that
            # the value matches the PostgreSQL syntax for numeric constants using a regex.
            # https://www.postgresql.org/docs/current/sql-syntax-lexical.html#SQL-SYNTAX-CONSTANTS-NUMERIC
            # 1, +1, -1, .1, 1., 1.0, 1E4, 1E+4, 1e-4, -4E2, -4.E5, .5e2, +123.456E+9 are all acceptable values.
            # This regex looks for an optional leading sign, then splits into two cases to handle both `xx(.(yy))`
            # and `.yy`, then accepts an optional exponent with an optional leading sign.
            models.CheckConstraint(
                check=~Q(type=MetaType.Numeric) | Q(value__iregex=r"^[+-]?(?:\d+(?:\.\d*)?|\.\d+)(?:E[+-]?\d+)?$"),
                name="metadata_numeric_values"
            ),
            # There can be a worker run ID only if there is a worker version ID,
            # but there can be a worker version ID without a worker run ID (backwards compatibility).
            # In other words, either the worker run ID is null, or the worker version ID is not null.
            models.CheckConstraint(
                check=Q(worker_version_id__isnull=False) | Q(worker_run_id__isnull=True),
                name="metadata_worker_run_requires_worker_version",
            )
        ]
        indexes = [
            models.Index(
                "type",
                name="metadata_type_index",
            ),
            models.Index(
                Cast("value", output_field=models.FloatField()),
                condition=Q(type=MetaType.Numeric),
                name="metadata_numeric_index",
            ),
        ]

    @property
    def raw_dates(self):
        if self.type != MetaType.Date:
            return
        return self.value

    def clean(self):
        if self.entity is None or self.element is None:
            return
        if self.entity.corpus != self.element.corpus:
            raise ValidationError(f"Entity's corpus {self.entity.corpus} is different from the expected corpus {self.element.corpus}")

    def save(self, *args, **kwargs):
        self.clean()
        super().save(*args, **kwargs)


class Selection(models.Model):
    id = models.UUIDField(default=uuid.uuid4, primary_key=True, editable=False)
    element = models.ForeignKey(Element, related_name="selections", on_delete=models.CASCADE)
    user = models.ForeignKey(settings.AUTH_USER_MODEL, related_name="selections", on_delete=models.CASCADE)

    class Meta:
        unique_together = (
            ("element", "user"),
        )


class CorpusExportState(Enum):
    Created = "created"
    Running = "running"
    Failed = "failed"
    Done = "done"


class CorpusExport(S3FileMixin, IndexableModel):
    corpus = models.ForeignKey(Corpus, related_name="exports", on_delete=models.CASCADE)
    user = models.ForeignKey(settings.AUTH_USER_MODEL, related_name="exports", on_delete=models.CASCADE)
    state = EnumField(CorpusExportState, max_length=10, default=CorpusExportState.Created)
    source = models.CharField(max_length=50, default="default", choices=[(source, source) for source in settings.EXPORT_SOURCES])

    s3_bucket = settings.AWS_EXPORT_BUCKET

    @property
    def s3_key(self) -> str:
        return str(self.id)

    def is_deletable(self, user) -> bool:
        """
        Whether or not the request user can delete this export
        """
        if user.is_anonymous or getattr(user, "is_agent", False):
            return False

        if user.is_admin:
            return True

        from arkindex.users.utils import get_max_level
        level = get_max_level(user, self.corpus)

        return level is not None and ((level >= Role.Admin.value) or (level >= Role.Contributor.value and self.user_id == user.id))

    def start(self):
        from arkindex.project.triggers import export_corpus
        assert self.state == CorpusExportState.Created, "This export has already been started"
        export_corpus(self)
