from unittest.mock import patch

from django.urls import reverse
from rest_framework import status

from arkindex.documents.models import Corpus
from arkindex.process.models import CorpusWorkerVersion, Worker, WorkerConfiguration
from arkindex.project.tests import FixtureAPITestCase
from arkindex.training.models import Model, ModelVersionState


class TestCorpusWorkerVersions(FixtureAPITestCase):

    @classmethod
    def setUpTestData(cls):
        super().setUpTestData()
        cls.worker_reco = Worker.objects.get(slug="reco")
        cls.worker_dla = Worker.objects.get(slug="dla")

        cls.version_1 = cls.worker_reco.versions.get()
        cls.version_2 = cls.worker_dla.versions.get()

        cls.model = Model.objects.create(
            name="Generic model",
            public=False,
        )
        cls.model_version = cls.model.versions.create(
            state=ModelVersionState.Available,
            hash="42",
            archive_hash="42",
            size=1337,
        )

    def test_no_login_public(self):
        self.assertTrue(self.corpus.public)
        # Use multiple versions to ensure we don't get duplicated queries
        corpus_worker_version_1 = self.corpus.worker_version_cache.create(worker_version=self.version_1)
        corpus_worker_version_2 = self.corpus.worker_version_cache.create(worker_version=self.version_2)

        with self.assertNumQueries(3):
            response = self.client.get(reverse("api:corpus-versions", kwargs={"pk": self.corpus.id}))
            self.assertEqual(response.status_code, status.HTTP_200_OK)

        self.assertDictEqual(response.json(), {
            "count": 2,
            "number": 1,
            "previous": None,
            "next": None,
            "results": [{
                "id": str(corpus_worker_version_2.id),
                "worker_configuration": None,
                "model_version": None,
                "worker_version": {
                    "id": str(self.version_2.id),
                    "configuration": {"test": 42},
                    "revision_url": None,
                    "version": 1,
                    "tag": None,
                    "branch": None,
                    "gpu_usage": "disabled",
                    "model_usage": "disabled",
                    "docker_image_iid": self.version_2.docker_image_iid,
                    "state": "available",
                    "worker": {
                        "id": str(self.worker_dla.id),
                        "name": "Document layout analyser",
                        "type": "dla",
                        "slug": "dla",
                        "description": "",
                        "archived": False,
                        "repository_url": None,
                    },
                    "created": "2020-02-02T01:23:45.678000Z",
                },
            }, {
                "id": str(corpus_worker_version_1.id),
                "worker_configuration": None,
                "model_version": None,
                "worker_version": {
                    "id": str(self.version_1.id),
                    "configuration": {"test": 42},
                    "revision_url": None,
                    "version": 1,
                    "tag": None,
                    "branch": None,
                    "gpu_usage": "disabled",
                    "model_usage": "disabled",
                    "docker_image_iid": self.version_1.docker_image_iid,
                    "state": "available",
                    "worker": {
                        "id": str(self.worker_reco.id),
                        "name": "Recognizer",
                        "type": "recognizer",
                        "slug": "reco",
                        "description": "",
                        "archived": False,
                        "repository_url": None,
                    },
                    "created": "2020-02-02T01:23:45.678000Z",
                },
            }]
        })

    @patch("arkindex.users.managers.BaseACLManager.filter_rights", return_value=Corpus.objects.filter(public=True))
    def test_no_login_private(self, filter_rights_mock):
        self.corpus.public = False
        self.corpus.save()
        self.corpus.worker_versions.set([self.version_1, self.version_2])

        with self.assertNumQueries(1):
            response = self.client.get(reverse("api:corpus-versions", kwargs={"pk": self.corpus.id}))
            self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

        self.assertDictEqual(response.json(), {"detail": "No Corpus matches the given query."})

    def test_not_verified(self):
        self.user.verified_email = False
        self.user.save()
        self.client.force_login(self.user)
        # Use multiple versions to ensure we don't get duplicated queries
        corpus_worker_version_1 = self.corpus.worker_version_cache.create(worker_version=self.version_1)
        corpus_worker_version_2 = self.corpus.worker_version_cache.create(worker_version=self.version_2)

        with self.assertNumQueries(5):
            response = self.client.get(reverse("api:corpus-versions", kwargs={"pk": self.corpus.id}))
            self.assertEqual(response.status_code, status.HTTP_200_OK)

        self.assertDictEqual(response.json(), {
            "count": 2,
            "number": 1,
            "previous": None,
            "next": None,
            "results": [{
                "id": str(corpus_worker_version_2.id),
                "worker_configuration": None,
                "model_version": None,
                "worker_version": {
                    "id": str(self.version_2.id),
                    "configuration": {"test": 42},
                    "revision_url": None,
                    "version": 1,
                    "tag": None,
                    "branch": None,
                    "gpu_usage": "disabled",
                    "model_usage": "disabled",
                    "docker_image_iid": self.version_2.docker_image_iid,
                    "state": "available",
                    "worker": {
                        "id": str(self.worker_dla.id),
                        "name": "Document layout analyser",
                        "type": "dla",
                        "slug": "dla",
                        "description": "",
                        "archived": False,
                        "repository_url": None,
                    },
                    "created": "2020-02-02T01:23:45.678000Z",
                },
            }, {
                "id": str(corpus_worker_version_1.id),
                "worker_configuration": None,
                "model_version": None,
                "worker_version": {
                    "id": str(self.version_1.id),
                    "configuration": {"test": 42},
                    "revision_url": None,
                    "version": 1,
                    "tag": None,
                    "branch": None,
                    "gpu_usage": "disabled",
                    "model_usage": "disabled",
                    "docker_image_iid": self.version_2.docker_image_iid,
                    "state": "available",
                    "worker": {
                        "id": str(self.worker_reco.id),
                        "name": "Recognizer",
                        "type": "recognizer",
                        "slug": "reco",
                        "description": "",
                        "archived": False,
                        "repository_url": None,
                    },
                    "created": "2020-02-02T01:23:45.678000Z",
                },
            }]
        })

    def test_list(self):
        self.client.force_login(self.user)
        # Use multiple versions with a model and worker configuration to ensure we don't get duplicated queries
        conf_1 = WorkerConfiguration.objects.create(name="test", configuration={"num": 1337}, worker=self.worker_reco)
        corpus_worker_version_1 = CorpusWorkerVersion.objects.create(
            corpus=self.corpus,
            worker_version=self.version_1,
            model_version=self.model_version,
            worker_configuration=conf_1,
        )
        conf_2 = WorkerConfiguration.objects.create(name="test 2", configuration={"value": "leet"}, worker=self.worker_dla)
        corpus_worker_version_2 = CorpusWorkerVersion.objects.create(
            corpus=self.corpus,
            worker_version=self.version_2,
            model_version=self.model_version,
            worker_configuration=conf_2,
        )

        with self.assertNumQueries(5):
            response = self.client.get(reverse("api:corpus-versions", kwargs={"pk": self.corpus.id}))
            self.assertEqual(response.status_code, status.HTTP_200_OK)

        self.assertDictEqual(response.json(), {
            "count": 2,
            "number": 1,
            "previous": None,
            "next": None,
            "results": [{
                "id": str(corpus_worker_version_2.id),
                "worker_version": {
                    "id": str(self.version_2.id),
                    "configuration": {"test": 42},
                    "revision_url": None,
                    "version": 1,
                    "tag": None,
                    "branch": None,
                    "gpu_usage": "disabled",
                    "model_usage": "disabled",
                    "docker_image_iid": self.version_2.docker_image_iid,
                    "state": "available",
                    "worker": {
                        "id": str(self.worker_dla.id),
                        "name": "Document layout analyser",
                        "type": "dla",
                        "slug": "dla",
                        "description": "",
                        "archived": False,
                        "repository_url": None,
                    },
                    "created": "2020-02-02T01:23:45.678000Z",
                },
                "worker_configuration": {
                    "id": str(conf_2.id),
                    "archived": False,
                    "name": "test 2",
                    "configuration": {"value": "leet"},
                },
                "model_version": {
                    "id": str(self.model_version.id),
                    "configuration": {},
                    "model": {
                        "id": str(self.model.id),
                        "name": "Generic model"
                    },
                    "size": 1337,
                    "state": "available",
                    "tag": None,
                },
            }, {
                "id": str(corpus_worker_version_1.id),
                "worker_version": {
                    "id": str(self.version_1.id),
                    "configuration": {"test": 42},
                    "revision_url": None,
                    "version": 1,
                    "tag": None,
                    "branch": None,
                    "gpu_usage": "disabled",
                    "model_usage": "disabled",
                    "docker_image_iid": self.version_1.docker_image_iid,
                    "state": "available",
                    "worker": {
                        "id": str(self.worker_reco.id),
                        "name": "Recognizer",
                        "type": "recognizer",
                        "slug": "reco",
                        "description": "",
                        "archived": False,
                        "repository_url": None,
                    },
                    "created": "2020-02-02T01:23:45.678000Z",
                },
                "worker_configuration": {
                    "id": str(conf_1.id),
                    "archived": False,
                    "name": "test",
                    "configuration": {"num": 1337},
                },
                "model_version": {
                    "id": str(self.model_version.id),
                    "configuration": {},
                    "model": {
                        "id": str(self.model.id),
                        "name": "Generic model"
                    },
                    "size": 1337,
                    "state": "available",
                    "tag": None,
                },
            }],
        })

    def test_no_right(self):
        self.user.rights.all().delete()
        self.client.force_login(self.user)
        corpus_worker_version = self.corpus.worker_version_cache.create(worker_version=self.version_1)

        with self.assertNumQueries(5):
            response = self.client.get(reverse("api:corpus-versions", kwargs={"pk": self.corpus.id}))
            self.assertEqual(response.status_code, status.HTTP_200_OK)

        self.assertDictEqual(response.json(), {
            "count": 1,
            "number": 1,
            "previous": None,
            "next": None,
            "results": [{
                "id": str(corpus_worker_version.id),
                "worker_configuration": None,
                "model_version": None,
                "worker_version": {
                    "id": str(self.version_1.id),
                    "configuration": {"test": 42},
                    "revision_url": None,
                    "version": 1,
                    "tag": None,
                    "branch": None,
                    "gpu_usage": "disabled",
                    "model_usage": "disabled",
                    "docker_image_iid": self.version_1.docker_image_iid,
                    "state": "available",
                    "worker": {
                        "id": str(self.worker_reco.id),
                        "name": "Recognizer",
                        "type": "recognizer",
                        "slug": "reco",
                        "description": "",
                        "archived": False,
                        "repository_url": None,
                    },
                    "created": "2020-02-02T01:23:45.678000Z",
                }
            }]
        })
